
module pid_controller_boost_converter(rst,clk,Kp,Ki,Kd,v_adc,u);

parameter e_nu=5;
parameter dpwm_n= 4;
parameter ddpwm_n= 4;
parameter kp_n=7;
parameter kp_q=7;
parameter kd_n=12;
parameter kd_q=6;
parameter ki_n=11;
parameter ki_q=12;
parameter up_n=11;
parameter ud_n=17;
parameter ui_n=21;
parameter wi_n=15;


parameter e_n=14; //adc full width
input clk;
wire clkd;
input rst;
wire signed [e_n-1:0] e_comp;
wire signed [e_nu-1:0] e;
reg signed [e_nu-1:0] ee;
input signed [kp_n-1:0] Kp;  //[kp_n,-kp_q]
input signed [ki_n-1:0] Ki; //[ki_n,-kp_q]
input signed [kd_n-1:0] Kd; //[kd_n,-kd_q]
input signed [e_n-1:0] v_adc; //[e_n,0]

///////  Error voltage /////////////
assign e_comp=14'b00111111111111-v_adc;   //comparator for reference voltage vref_digital=2084 for vref=12/47.16 and ADC input range=2vpp  ----00100000100100
assign e=e_comp[e_n-1:e_n-e_nu];  //final error voltage for our ADC-bits


output reg u; //[1,0]
reg signed [e_nu-1:0] e1; //[e_n,0]
wire signed [up_n-1:0] up; //[up_n,-kp_q]
wire signed [up_n+5:0] upp; //[up_n+6,-kp_q]
wire signed[ui_n-1:0] ui; //[ui_n,-ki_q]
reg signed [ui_n-1:0] ui1; //[ui_n,-ki_q]
wire signed [wi_n-1:0] wi; //[wi_n,-ki_q]
wire signed [wi_n+5:0] wii; //[wi_n+6,-ki_q]
wire signed [ud_n-1:0] ud; //[ud_n,-kd_q]
wire signed [ud_n+5:0] udd; //[ud_n+6,-kd_q]
wire signed [e_nu:0] wd; //[e_n+1,0]
wire signed [ud_n+5+(ki_q-kd_q):0] upd; //[ud_n+6+(ki_q-kd_q),-ki_q]
wire signed [ui_n-1:0] upid; //[ui_n,-ki_q]
wire signed [ui_n-1-ki_q:0] ux; //[ui_n-ki_q,0]
wire unsigned [dpwm_n+ddpwm_n-1:0]uh; //[dpwm_n+ddpwm_n,0]
reg unsigned [dpwm_n+ddpwm_n-1:0]uhh; //[dpwm_n+ddpwm_n,0]
reg [dpwm_n-1:0] count;
reg [ddpwm_n-1:0] countd;
wire ddpm_out;
wire [ddpwm_n-1:0] u_lsb; //[ddpwm_n,0]


initial
begin
ui1<={(ui_n){1'b0}};
count<=0;
e1<=0;
ee<=e;
countd<=0;
uhh<=0;
end
always@(negedge clk)
if(rst)
begin
ui1<={(ui_n){1'b0}};
count<=0;
e1<=0;
ee<=e;
countd<=0;
uhh<=0;
end

//********* Clock Divider********//
always@(posedge clk)
begin
if(count==2**dpwm_n-1)
count<=0;
else
count<=count+1;
end
assign clkd=~count[dpwm_n-1];
//********************************//

//****************************************
///Combinationalpart
//****************************************
saturated_multiplier Kp_mult (ee,Kp,up,);
defparam Kp_mult.n=e_nu,Kp_mult.p=kp_n,Kp_mult.m=up_n;
saturated_multiplier Kp_mult4 (up,7'b0011000,upp,);
defparam Kp_mult4.n=up_n,Kp_mult4.p=7,Kp_mult4.m=up_n+7-1;
saturated_multiplier Ki_mult(ee,Ki,wi,);
defparam Ki_mult.n=e_nu,Ki_mult.p=ki_n,Ki_mult.m=wi_n;
saturated_multiplier Kp_mult5 (wi,7'b0011000,wii,);  //0101111
defparam Kp_mult5.n=wi_n,Kp_mult5.p=7,Kp_mult5.m=wi_n+7-1;
saturated_adder Ki_add(ui1,wii,ui,,1'b1);
defparam Ki_add.n=ui_n,Ki_add.p=wi_n+7-1,Ki_add.m=ui_n;
saturated_adder Kd_sub(ee,e1,wd,,1'b0);
defparam Kd_sub.n=e_nu,Kd_sub.p=e_nu,Kd_sub.m=e_nu+1;
saturated_multiplier Kd_mult(wd,Kd,ud,);
defparam Kd_mult.n=e_nu+1,Kd_mult.p=kd_n,Kd_mult.m=ud_n;
saturated_multiplier Kp_mult6 (ud,7'b0011000,udd,);
defparam Kp_mult6.n=ud_n,Kp_mult6.p=7,Kp_mult6.m=ud_n+7-1;
saturated_adder upd_add({upp,{(ki_q-kp_q){1'b0}}},{udd,{(ki_q-kd_q){1'b0}}},upd,,1'b1);
defparam upd_add.n=up_n+7-1+(ki_q-kp_q),upd_add.p=ud_n+7-1+(ki_q-kd_q),upd_add.m=ud_n+7-1+(ki_q-kd_q);
saturated_adder upid_add(upd,ui,upid,,1'b1);
defparam upid_add.n=ud_n+7-1+(ki_q-kd_q),upid_add.p=ui_n,upid_add.m=ui_n;
assign ux=upid[ui_n-1:ki_q];
assign uh=(ux[ui_n-1-ki_q]==1'b1)?{(dpwm_n+ddpwm_n){1'b0}}:ux>{(dpwm_n+ddpwm_n){1'b1}}?{(dpwm_n+ddpwm_n){1'b1}}:ux[dpwm_n+ddpwm_n-1:0];
assign u_lsb=uhh[ddpwm_n-1:0]; 
ddpm uddp(uhh[ddpwm_n-1:0],ddpm_out,countd);
defparam uddp.RX=ddpwm_n;
//****************************************
//Synchronization part
//****************************************
always@(posedge clkd)
uhh<=uh;


always@(posedge clk)
u <= (uhh[dpwm_n+ddpwm_n-1:ddpwm_n]+ddpm_out)>count;
//****************************************
//Sequential part
//****************************************
always@(posedge clkd)
begin
e1<=ee;
ee<=e;
ui1<=ui;
if(countd==2**ddpwm_n-1)
countd<=0;
else
countd<=countd+1;
end
endmodule



module ddpm(
in,
ddpm_out,
count
);

parameter RX;

input [RX-1:0] in;
input [RX-1:0] count;
output ddpm_out;


integer  i;
integer  index;

reg  [RX:0] data;
reg  ddpm_out;


always @ (*) 

begin

index=RX;

i=0;

for (i=RX-1; i>=0; i=i-1) begin

	if (count[i]==1) begin

		index=i;

    end	 		 	 

end

index=RX-index;
data<= {in, 1'b0};
ddpm_out <= data[index];
  
end
endmodule

 

 
module saturated_adder(x,y,z,OV,op);
function integer max;
input integer left,right;
if(left>right)
max=left;
else
max=right;
endfunction
parameter n;
parameter p;
parameter m;//Assumingm<=max(n,p)+1
parameter mx=max(n,p)+1;
input signed [n-1:0]x;
input signed [p-1:0]y;
output reg signed [m-1:0] z;
input op;
output reg OV;
wire signed[n:0]xx={x[n-1],x};
wire signed[p:0]yx={y[p-1],y};
wire signed[mx-1:0]zx;
assign zx=(op==1'b1)?xx+yx:xx-yx;
reg temp;
integer I;
always@(zx)
begin
temp=1'b0;
for(I=m;I<=mx-1;I=I+1)
begin
if((zx[I]^zx[m-1])==1'b1)
temp=1'b1;
end
OV=temp;
end
always@(OV,zx)
case(OV)
1'b0:z=zx[m-1:0];
1'b1:
begin
if(zx[mx-1]==1'b0)
z={1'b0,{(m-1){1'b1}}};
else
z={1'b1,{(m-1){1'b0}}};
end
endcase
endmodule


module saturated_multiplier(x,y,z,OV);
parameter n;
parameter p;
parameter m;//Assumingm<=n+p
input signed [n-1:0] x;//[n,q]
input signed[p-1:0]y;//[p,l]
output reg [m-1:0]z;//[m,q+l]
output reg OV;
wire signed [n+p-1:0]zx;
assign zx=x*y;
reg temp;
integer I;
always@(zx)
begin
temp=1'b0;
for(I=m;I<=n+p-1;I=I+1)
begin
if((zx[I]^zx[m-1])==1'b1)
temp=1'b1;
end
OV=temp;
end
always@(OV,zx)
case(OV)
1'b0:z=zx[m-1:0];
1'b1:
begin
if(zx[n+p-1]==1'b0)
z={1'b0,{(m-1){1'b1}}};
else
z={1'b1,{(m-1){1'b0}}};
end
endcase
endmodule





